@@ -10,9 +10,11 @@
 #include "fileindexerconfig.h"
 #include "pendingfilequeue.h"
 
-#include <QTest>
+#include <QSaveFile>
 #include <QSignalSpy>
 #include <QTemporaryDir>
+#include <QTest>
+
 #include <KFileMetaData/UserMetaData>
 
 namespace Baloo {
@@ -26,6 +28,7 @@ private Q_SLOTS:
     void testDirCreation();
     void testConfigChange();
     void testFileMoved();
+    void testFileReplaced();
 
     void init()
     {
@@ -51,7 +54,9 @@ using namespace Baloo;
 namespace {
     bool createFile(const QString& fileUrl) {
         QFile f1(fileUrl);
-        f1.open(QIODevice::WriteOnly);
+        if (!f1.open(QIODevice::WriteOnly)) {
+            return false;
+        }
         f1.close();
         return QFile::exists(fileUrl);
     }
@@ -63,10 +68,9 @@ namespace {
 
     void modifyFile(const QString& fileUrl) {
         QFile f1(fileUrl);
-        f1.open(QIODevice::Append | QIODevice::Text);
+        QVERIFY(f1.open(QIODevice::Append));
 
-        QTextStream stream(&f1);
-        stream << "1";
+        QCOMPARE(f1.write(QByteArray("abcd", 4)), 4);
     }
 }
 
@@ -328,12 +332,73 @@ void FileWatchTest::testFileMoved()
     // Move the file
     QVERIFY(QFile::rename(fileUrl, fileDestUrl));
 
-    QVERIFY(spyIndexFileRemoved.wait());
+    // Moving an unindexd file creates two events, in order:
+    // 1) A cooked "removed" event for the destination, for a potentially overwritten file
+    // 2) A "new" event for the moved file, with the new path
+    QVERIFY(spyIndexNew.wait());
+
     QCOMPARE(spyIndexNew.count(), 1);
     QCOMPARE(spyIndexNew.takeFirst().at(0), fileDestUrl);
     QCOMPARE(spyIndexModified.count(), 0);
     QCOMPARE(spyIndexFileRemoved.count(), 1);
-    QCOMPARE(spyIndexFileRemoved.takeFirst().at(0), fileDestUrl); // called to clean dest metadata
+
+#ifdef Q_OS_FREEBSD
+    QEXPECT_FAIL("", "FreeBSD fails to report MOVE_FROM/MOVE_TO pair (incorrect cookie).", Abort);
+#endif
+    QCOMPARE(spyIndexFileRemoved.takeFirst().at(0), fileDestUrl);
+}
+
+void FileWatchTest::testFileReplaced()
+{
+    QCOMPARE(m_db->open(Database::CreateDatabase), Database::OpenResult::Success);
+
+    Test::writeIndexerConfig({m_tmpDir->path()}, {});
+    FileIndexerConfig config;
+
+    FileWatch fileWatch(m_db.get(), &config);
+    fileWatch.m_pendingFileQueue->setMaximumTimeout(0);
+    fileWatch.m_pendingFileQueue->setMinimumTimeout(0);
+    fileWatch.m_pendingFileQueue->setTrackingTime(0);
+
+    QSignalSpy spy(&fileWatch, &FileWatch::installedWatches);
+    QVERIFY(spy.isValid());
+
+    fileWatch.updateIndexedFoldersWatches();
+    QVERIFY(spy.wait());
+    QCOMPARE(spy.count(), 1);
+
+    QSignalSpy spyIndexNew(&fileWatch, &FileWatch::indexNewFile);
+    QSignalSpy spyIndexModified(&fileWatch, &FileWatch::indexModifiedFile);
+    QSignalSpy spyIndexFileRemoved(&fileWatch, &FileWatch::fileRemoved);
+
+    QVERIFY(spyIndexNew.isValid());
+    QVERIFY(spyIndexModified.isValid());
+    QVERIFY(spyIndexFileRemoved.isValid());
+
+    // Create a file and see if it is indexed
+    QString fileUrl(m_tmpDir->path() + QStringLiteral("/t1"));
+    QVERIFY(createFile(fileUrl));
+
+    QVERIFY(spyIndexNew.wait());
+    QCOMPARE(spyIndexNew.count(), 1);
+    QCOMPARE(spyIndexModified.count(), 0);
+    QCOMPARE(spyIndexFileRemoved.count(), 0);
+    QCOMPARE(spyIndexNew.takeFirst().at(0), fileUrl);
+
+    // Overwrite atomically
+    {
+        QSaveFile t(fileUrl);
+        QVERIFY(t.open(QIODeviceBase::WriteOnly));
+        QVERIFY(t.commit());
+    }
+
+    QVERIFY(spyIndexNew.wait());
+
+    QCOMPARE(spyIndexNew.count(), 1);
+    QCOMPARE(spyIndexNew.takeFirst().at(0), fileUrl);
+    QCOMPARE(spyIndexModified.count(), 0);
+    QCOMPARE(spyIndexFileRemoved.count(), 1);
+    QCOMPARE(spyIndexFileRemoved.takeFirst().at(0), fileUrl); // cooked event to signal overwriting the old file
 }
 
 QTEST_MAIN(FileWatchTest)
